@@ -1,22 +1,20 @@
-"""Split a PDF document based on the locations of barcodes
-"""
-
-# Author: David A. Thompson
-
 import argparse
 import imp
 import json
+import math
 import os
 import os.path
 import re
+import shutil
 import signal
 import sys
 import tempfile
-import traceback
 import uuid
 
 import logging
 import logging.config
+
+
 # define LG as the generic logger *prior* to loading any
 # pdfxcb-specific modules
 lg = logging
@@ -24,33 +22,10 @@ lg = logging
 json_log_format = '%(message)s'
 lg.basicConfig(format=json_log_format,stream=sys.stdout)
 
-# internal modules
-from . import barScan
-#import bubbles
-#import deskew
-#import exceptions
-from . import json1
-from . import pdf
-from . import util
 
-
-# bubbles.py, deskew.py, feature_detect.py, and util.py use numpy
-# try:
-#     imp.find_module('numpy')
-# except ImportError:
-#     msg = json1.json_msg_module_not_accessible('numpy')
-#     lg.error(msg)
-#     lg.info(json1.json_last_log_msg())
-#     sys.exit(msg)
-
-try:
-    imp.find_module('PyPDF2')
-except ImportError:
-    msg = json1.json_msg_module_not_accessible('PyPDF2')
-    lg.error(msg)
-    lg.info(json1.json_last_log_msg())
-    sys.exit(msg)
-import PyPDF2
+import pdfxcb.barScan as barScan
+import pdfxcb.json1 as json1
+import pdfxcb.pdf as pdf
 
 
 # handle external signals requesting termination
@@ -71,9 +46,11 @@ signal.signal(signal.SIGTERM, signal_handler)
 #
 def locate_cover_sheets (png_file_tuples,containing_dir,match_re,scan_region):
     """
-    Given the list of files specified by PNG_FILE_TUPLES (tuples where the first member specifies the name of the PNG file) and CONTAINING_DIR,
-    identify those files containing a barcode. Return multiple values: a list of the
-    corresponding barcodes and a list of the corresponding indices.
+    Given the list of files specified by PNG_FILE_TUPLES (a set of
+    tuples where the first member of each tuple specifies the name of
+    the PNG file) and CONTAINING_DIR, identify those files containing
+    a barcode. Return multiple values: a list of the corresponding
+    barcodes and a list of the corresponding indices.
     """
     barcodes = []
     indices = []
@@ -82,7 +59,12 @@ def locate_cover_sheets (png_file_tuples,containing_dir,match_re,scan_region):
     i_max = len(png_file_tuples)
     while (i<i_max):
         # log progress by default (otherwise, this can be a long period of silence...)
-        lg.info(json1.json_progress("looking for barcode on " + str(i) + " of " + str(i_max) + " PNG files"))
+        lg.info(
+            json1.json_progress(
+                f'looking for barcode on {i} of {i_max} PNG files')
+            )
+        lg.info(containing_dir)
+        lg.info(png_file_tuples[i][0])
         image_file_spec = os.path.join(containing_dir,png_file_tuples[i][0])
         lg.debug(image_file_spec)
         maybe_barcode = barScan.barcodeScan(
@@ -104,37 +86,59 @@ def locate_cover_sheets (png_file_tuples,containing_dir,match_re,scan_region):
 
 def executable_sanity_checks (executables):
     """
-    Check for availability of executables specified in the list of strings EXECUTABLES.
+    Check for availability of executables specified in the list of
+    strings EXECUTABLES.
     """
-    lg.debug(util)
     for executable_spec in executables:
-        if not util.which(executable_spec):
+        if not shutil.which(executable_spec):
             msg = json1.json_msg_executable_not_accessible(executable_spec)
             lg.error(msg)
             lg.info(json1.json_last_log_msg())
             sys.exit(msg)
 
-def generate_output_file_names(cover_sheet_barcodes,cover_sheet_indices,output_dir):
+def generate_output_file_names(cover_sheet_barcodes,
+                               cover_sheet_indices,
+                               output_dir):
     file_names = []
     for cover_sheet_barcode,cover_sheet_index in zip(cover_sheet_barcodes,cover_sheet_indices):
         cover_sheet_index_as_string = str.format("{0:0>03d}", cover_sheet_index)
         version = -1
         # sanity check on version (completely arbitrary at this point)
         max_version = 99
-        while (not version >= max_version and (version < 0 or os.path.exists(path))):
+        while (not version >= max_version and
+               (version < 0 or os.path.exists(path))):
             version = version + 1
-            file_name = cover_sheet_barcode + "-" + cover_sheet_index_as_string + "-" + str(version) + ".pdf"
+            file_name = f'{cover_sheet_barcode}-{cover_sheet_index_as_string}-{version}.pdf'
             path = os.path.join(output_dir,file_name)
         file_names.append(path)
     return file_names
 
-def generate_page_ranges(cover_sheet_indices,png_file_page_number_tuples,number_of_pages):
+def generate_output_file_names_split_after(page_ranges,output_dir):
+    file_names = []
+    for page_range in page_ranges:
+        page_range_as_string = str.format("{0:0>03d}", page_range[0]) + "-"+str.format("{0:0>03d}",page_range[1])
+        version = -1
+        # sanity check on version (completely arbitrary at this point)
+        max_version = 99
+        while (not version >= max_version and
+               (version < 0 or os.path.exists(path))):
+            version = version + 1
+            file_name = f'{page_range_as_string}-{version}.pdf'
+            path = os.path.join(output_dir,file_name)
+        file_names.append(path)
+    return file_names
+
+def generate_page_ranges(cover_sheet_indices,
+                         png_file_page_number_tuples,
+                         number_of_pages):
     """
-    Calling code must guarantee that tuples in
-    PNG_FILE_PAGE_NUMBER_TUPLES are ordered (ascending) with respect
-    to page numbers. COVER_SHEET_INDICES is an array of integers, each
-    an index value identifying a member of PNG_FILE_PAGE_NUMBER_TUPLES
-    which corresponds to a cover sheet.
+    Return a list of tuples. COVER_SHEET_INDICES is an array of
+    integers, each an index value identifying a member of
+    PNG_FILE_PAGE_NUMBER_TUPLES which corresponds to a cover sheet.
+    See INVOKE_PDFIMAGES_ON docstring for more on
+    PNG_FILE_PAGE_NUMBER_TUPLES. Calling code must guarantee that
+    tuples in PNG_FILE_PAGE_NUMBER_TUPLES are ordered (ascending) with
+    respect to page numbers.
     """
     # to capture last set of pages, tag on an imaginary cover sheet at the end
     cover_sheet_indices.append(
@@ -148,21 +152,47 @@ def generate_page_ranges(cover_sheet_indices,png_file_page_number_tuples,number_
              png_file_page_number_tuples[next_cover_sheet_index][1]-1))
     return page_ranges
 
-def sanity_checks (dirs,files):
-    lg.debug(files)
+def generate_page_ranges_split_after(split_after,number_of_pages):
+    """
+    Return a list of tuples.
+    """
+    number_of_docs = int(math.ceil(number_of_pages/split_after))
+    page_ranges = [(n*split_after-(split_after-1),n*split_after) for n in range(1,number_of_docs+1)]
+    # adjust the last tuple since the number of pages may not be a
+    # multiple of SPLIT_AFTER (the last set of pages may be smaller
+    # than split_after pages)
+    last_tuple = page_ranges[len(page_ranges)-1]
+    page_ranges[len(page_ranges)-1] = (last_tuple[0],number_of_pages)
+    return page_ranges
+
+def pdfxcb_sanity_checks (output_dir,pdf_file_spec,rasterize_p,region):
+    # file and dir sanity checks
+    if (not output_dir):
+        sys.exit("The output directory must be specified.")
+    directory_sanity_checks ([output_dir],True)
+    file_sanity_checks ([pdf_file_spec],True)
+    # region/rasterize sanity check
+    if (not rasterize_p and region):
+        sys.exit("If REGION is specified, then RASTERIZE_P should be true.")
+    # executables sanity check
     required_executables = [
         'gs'
         #'pdftoppm'
     ]
     executable_sanity_checks(required_executables)
-    directory_sanity_checks (dirs,True)
-    file_sanity_checks (files,True)
     required_modules = [
+        'PyPDF2'
         #'cv2'
     ]
     module_sanity_checks (required_modules,True)
 
-def pdfxcb (pdf_file_spec,output_dir,match_re,rasterize_p):
+def file_and_dir_sanity_checks (dirs,files):
+    directory_sanity_checks (dirs,True)
+    file_sanity_checks (files,True)
+
+def pdfxcb (pdf_file_spec,output_dir,match_re,rasterize_p,region,
+            clean_up_png_files_p=True
+            ):
     """
     Given the file specified by PDF_FILE_SPEC, look for cover sheets
     and split the PDF at each coversheet. Name output file(s) based on
@@ -171,10 +201,12 @@ def pdfxcb (pdf_file_spec,output_dir,match_re,rasterize_p):
     unless the corresponding string matches the regex MATCH_RE. Use
     RASTERIZE_P = False if the PDF does not contain vector graphics
     but is solely bitmap data (e.g., the PDF was generated from a
-    scanned document).
+    scanned document). If REGION has the form [ float1, float2,
+    float3, float4 ], use the region specified by REGION when scanning
+    for a barcode or other indicator of a cover sheet.
     """
     global lg
-    sanity_checks([output_dir],[pdf_file_spec])
+    pdfxcb_sanity_checks(output_dir,pdf_file_spec,rasterize_p,region)
     # If confident that the PDF under analysis is derived from a scan
     # (i.e., contains only bitmap data), then the images embedded in
     # the PDF can be analyzed directly. If the PDF may contain vector
@@ -194,7 +226,7 @@ def pdfxcb (pdf_file_spec,output_dir,match_re,rasterize_p):
         png_file_page_number_tuples = split_pdf_to_png_files(pdf_file_spec,output_dir)
         # Once rasterized pages are generated, optionally scan for cue marks
         # CUE_INDICES = array where each member is an integer indicating index of member of png_file_page_number_tuples where the corresponding bitmap has a cue mark
-        # cue_indices = scan_for_cue_marks(png_file_page_number_tuples) <-- use urh_corner_mean w/reasonable threshold (10? 20? 50?) for "black" 
+        # cue_indices = scan_for_cue_marks(png_file_page_number_tuples) <-- use urh_corner_mean w/reasonable threshold (10? 20? 50?) for "black"
     else:
         # extract images directly from PDF
         png_file_page_number_tuples = invoke_pdfimages_on(pdf_file_spec,output_dir)
@@ -205,27 +237,29 @@ def pdfxcb (pdf_file_spec,output_dir,match_re,rasterize_p):
     #
     # locate cover sheets
     #
+    lg.info("Locating cover sheets")
     if rasterize_p:
         # possibilities:
         # 1. png files represent rasterized pages
-        scan_region = ([0,0,0.7,0.5])
+        if region:
+            scan_region = region
+        else:
+            scan_region = ([0,0,0.7,0.5])
     else:
         # 2. png files represent images from PDF (via pdfimages)
         scan_region = None # None is not treated as the equivalent of ([0,0,1,1]). ([0,0,1,1]) triggers cropping by barcodeScan.
     cover_sheet_barcodes, cover_sheet_indices = locate_cover_sheets(png_file_page_number_tuples,output_dir,match_re,scan_region)
     print(cover_sheet_barcodes)
-    lg.debug(cover_sheet_barcodes)
-    lg.debug(cover_sheet_indices)
-    # Setting to False supports debugging/development. This should be set to True in production.
-    clean_up_png_files = False # False # True
-    if clean_up_png_files:
+    if clean_up_png_files_p:
         for png_file_tuple in png_file_page_number_tuples:
             os.remove(os.path.join(output_dir,png_file_tuple[0]))
-    # write PDFs
-    pdf_length = pdf.pdf_number_of_pages(pdf_file_spec) # len(png_files) only works if PNGs are rasterized pages
-    page_ranges = generate_page_ranges(cover_sheet_indices,png_file_page_number_tuples,pdf_length)
-    output_file_names = generate_output_file_names(cover_sheet_barcodes,cover_sheet_indices,output_dir)
-    lg.debug(output_file_names)
+    pdf_length = pdf.pdf_number_of_pages(pdf_file_spec)
+    page_ranges = generate_page_ranges(cover_sheet_indices,
+                                       png_file_page_number_tuples,
+                                       pdf_length)
+    output_file_names = generate_output_file_names(cover_sheet_barcodes,
+                                                   cover_sheet_indices,
+                                                   output_dir)
     pdf.pdf_split(pdf_file_spec,output_file_names,page_ranges)
     lg.info(json1.json_msg(40,
              ['Analysis and burst completed'],
@@ -238,6 +272,31 @@ def pdfxcb (pdf_file_spec,output_dir,match_re,rasterize_p):
     ))
     return True
 
+def pdfxcb_split_after (pdf_file_spec,output_dir,split_after_n_pp):
+    """
+    Given the file specified by PDF_FILE_SPEC, split the PDF after
+    every SPLIT_AFTER pages. Name output file(s) based page ranges.
+    Write files to directory specified by OUTPUT_DIR. Return True.
+    """
+    global lg
+    file_and_dir_sanity_checks([output_dir],[pdf_file_spec])
+    pdf_length = pdf.pdf_number_of_pages(pdf_file_spec)
+    page_ranges = generate_page_ranges_split_after(split_after_n_pp,
+                                                   pdf_length)
+    output_file_names = generate_output_file_names_split_after(page_ranges,
+                                                               output_dir)
+    pdf.pdf_split(pdf_file_spec,output_file_names,page_ranges)
+    lg.info(json1.json_msg(40,
+             ['Analysis and burst completed'],
+             False,
+             files=output_file_names,
+             data={
+                 #'barcodes': cover_sheet_barcodes,
+                 #'indices': cover_sheet_indices
+             }
+    ))
+    return True
+
 def directory_sanity_check (directory_spec,exitp):
     if not os.path.isdir(directory_spec):
         lg.error(json1.json_file_not_found(directory_spec))
@@ -267,7 +326,7 @@ def invoke_pdfimages_on (pdf_file_spec,output_dir):
     to directory specified by OUTPUT_DIR.
 
     Returns a list of tuples where each tuple has the structure
-    (png_file,png_file_page_number) png_file_page_number is an
+    (png_file,png_file_page_number) where png_file is a string representing the file name and png_file_page_number is an
     integer. The list is an ordered sequence with respect to page
     number - low to high.
     """
@@ -279,7 +338,9 @@ def invoke_pdfimages_on (pdf_file_spec,output_dir):
             lg.error(json1.json_msg(108,[msg],False,files=[pdf_file_spec]))
             sys.exit(msg)
         else:
-            png_file_page_number_tuples = pdf.pdfimages(pdf_file_spec,output_dir)
+            lg.info("png_file_page_number_tuples 0")
+            png_file_page_number_tuples = pdf.pdfimages(pdf_file_spec,
+                                                        output_dir)
     except Exception as e:
         lg.debug(str(e))
         msg = json1.json_failed_to_convert_pdf(e,pdf_file_spec)
@@ -287,7 +348,6 @@ def invoke_pdfimages_on (pdf_file_spec,output_dir):
         lg.info(json1.json_last_log_msg())
         sys.exit(msg)
     else:
-        # Is it really import to log png files? (Need to dig them out of tuples...)
         lg.info(json1.json_pdf_to_pngs_success(pdf_file_spec,
                                                None #png_files
         ))
@@ -300,7 +360,6 @@ def module_sanity_checks (module_names,exitp):
 
 def module_sanity_check (module_name,exitp):
     """MODULE_NAME is a string"""
-    # check for presence of module which might not be installed/accessible
     try:
         imp.find_module(module_name)
     except ImportError:
@@ -314,7 +373,12 @@ def split_pdf_to_png_files (pdf_file_spec,output_dir):
     """
     Split the PDF file specified by PDF_FILE_SPEC into a series of
     files, each representing a single page as a PNG image. Write files
-    to directory specified by OUTPUT_DIR.
+    to the directory specified by OUTPUT_DIR.
+
+    Return a list of tuples where the first member of each tuple is a
+    string representing the file name and the second member of each
+    tuple is the corresponding page number (page numbering begins at
+    1).
     """
     png_files = None
     try:
@@ -359,11 +423,21 @@ def main():
     """Handle command-line invocation of pdfxcb.py."""
     global lg
     parser = argparse.ArgumentParser(description="This is pdfxcb")
+    # Split after every Nth page. If N is 3, split into pp 1-3, 4-6, 7-9, etc.
+    parser.add_argument("-e",
+                        help="A whole number. If specified, split after every Nth page, ignoring -m and -r.",
+                        action="store",
+                        default=0,
+                        dest="split_after_n_pp",
+                        type=int
+                        )
     parser.add_argument("-f",
-                        help="absolute path to log file",
+                        # FIXME: "busca.log" shouldn't be hardcoded -- set up a default_log_path var to handle this
+                        help="Absolute path to log file. If not specified, logs to busca.log.",
                         action="store",
                         dest="log_file",
                         type=str)
+    # FIXME: is this mandatory? If so, this should be explicit.
     parser.add_argument("-d",
                         help="absolute path to output directory",
                         action="store",
@@ -379,13 +453,24 @@ def main():
                         action="store",
                         dest="identifier",
                         type=str)
+    # These four values correspond to scan_region = ([0,0,0.7,0.5])
+    parser.add_argument("-r",
+                        help="specify region of the page to be evaluated",
+                        action="store",
+                        dest="region",
+                        nargs=4,
+                        type=float)
     parser.add_argument("-l",
                         help="integer between 0 (verbose) and 51 (terse) defining logging",
                         action="store",
                         dest="log_level",
                         type=int)
-    #parser.add_argument('-v', '--version', action='version', version=version.version)
-    parser.add_argument("input_files", help="an input (PDF) file",
+    # https://stackoverflow.com/questions/458550/standard-way-to-embed-version-into-python-package
+    parser.add_argument('-v', '--version', action='version', version="0.0.3")
+    parser.add_argument('--debug',
+                        help="do not clean up files used during processing"
+                        )
+    parser.add_argument("input_files", help="one or more input (PDF) files",
                         # keep nargs as we may want to accept multiple PDFs as input at some point
                         nargs=1,
                         type=str)
@@ -416,12 +501,17 @@ def main():
     file_handler.setFormatter(formatter)
     lg.getLogger().addHandler(file_handler)
     lg.getLogger().setLevel(log_level)
-    lg.debug("args: %s", args)
-    lg.debug("sys.argv: %s",sys.argv)
     if args.identifier:
         identifier = args.identifier
     else:
         identifier = str(uuid.uuid1())
+    rasterize_p = False
+    if args.region:
+        region = args.region
+        rasterize_p = True  # if region is specified, rasterize_p should be true
+        # lg.debug("args.region: %s", args.region)
+    else:
+        region = None
     # 1000[0-9][0-9][0-9]$ matches on tt user id
     match_re_string = args.match_re_string
     lg.debug(match_re_string)
@@ -431,17 +521,25 @@ def main():
     pdf_file_spec = args.input_files[0]
     lg.debug(pdf_file_spec)
     lg.info(json1.json_first_log_msg(identifier, files = [pdf_file_spec] ))
-    rasterize_p = False
     # generic debugging
     lg.debug(os.getcwd())         # current/working directory
     # might also want to import platform to get architecture, other details...
-    try:
-        pdfxcb(pdf_file_spec,args.output_dir,match_re,rasterize_p)
-    except Exception as e:
-        lg.error("Crash and burn")
-        lg.error(sys.exc_info()[0])
-        raise
-    lg.info(json1.json_last_log_msg())
+    if (args.split_after_n_pp > 0):
+        pdfxcb_split_after(pdf_file_spec,args.output_dir,args.split_after_n_pp)
+    else:
+        try:
+            pdfxcb(pdf_file_spec,
+                   args.output_dir,
+                   match_re,
+                   rasterize_p,
+                   region,
+                   not args.debug #clean_up_png_files_p
+                   )
+        except Exception as e:
+            lg.error("Crash and burn")
+            lg.error(sys.exc_info()[0])
+            raise
+        lg.info(json1.json_last_log_msg())
 
 if __name__ == "__main__":
     main()
